import bencher as bch


class ExampleFile(bch.ParametrizedSweep):
    content = bch.StringSweep(["entry1", "entry2", "entry3"])

    file_result = bch.ResultPath()

    def __call__(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)

        # this generates a unique filename and stores it in the cache directory
        filename = bch.gen_path(self.content, suffix=".txt")
        with open(filename, "w", encoding="utf-8") as text_file:
            text_file.write(f"content:{self.content}")
        self.file_result = filename
        return super().__call__()


def example_filepath(run_cfg: bch.BenchRunCfg = None, report: bch.BenchReport = None):
    bench = ExampleFile().to_bench(run_cfg, report)
    bench.plot_sweep()
    return bench


if __name__ == "__main__":
    example_filepath().report.show()



# from https://stackoverflow.com/questions/22989372/how-to-format-a-floating-number-to-maximum-fixed-width-in-python


class FormatFloat:
    def __init__(self, width=8):
        self.width = width
        self.maxnum = int("9" * (width - 1))  # 9999999
        self.minnum = -int("9" * (width - 2))  # -999999

    def __call__(self, x):
        # for small numbers
        # if -999,999 < given < 9,999,999:
        if self.minnum < x < self.maxnum:
            # o = f'{x:7}'
            o = f"{x:{self.width - 1}}"

            # converting int to float without adding zero
            if "." not in o:
                o += "."

            # float longer than 8 will need rounding to fit width
            elif len(o) > self.width:
                # output = str(round(x, 7 - str(x).index(".")))
                o = str(round(x, self.width - 1 - str(x).index(".")))
                if len(o) < self.width:
                    o += (self.width - len(o)) * "0"

        else:
            # for exponents
            # added a loop for super large numbers or negative as "-" is another char
            # Added max(max_char, 5) to account for max length of less
            #     than 5, was having too much fun
            # TODO can i come up with a threshold value for these up front,
            #     so that i dont have to do this calc for every value??
            for n in range(max(self.width, 5) - 5, 0, -1):
                fill = f".{n}e"
                o = f"{x:{fill}}".replace("+0", "+")

                # if all good stop looping
                if len(o) == self.width:
                    break
            else:
                raise ValueError(f"Number is too large to fit in {self.width} characters", x)
        return o



from collections import namedtuple
import xarray as xr
from sortedcontainers import SortedDict
import hashlib
import re
import math
from colorsys import hsv_to_rgb
from pathlib import Path
from uuid import uuid4
from functools import partial
from typing import Callable, Any, List, Tuple
import logging
import os
import tempfile
import shutil

import param
import numpy as np


def hmap_canonical_input(dic: dict) -> tuple:
    """From a dictionary of kwargs, return a hashable representation (tuple) that is always the same for the same inputs and retains the order of the input arguments.  e.g, {x=1,y=2} -> (1,2) and {y=2,x=1} -> (1,2).  This is used so that keywords arguments can be hashed and converted the the tuple keys that are used for holomaps

    Args:
        dic (dict): dictionary with keyword arguments and values in any order

    Returns:
        tuple: values of the dictionary always in the same order and hashable
    """

    function_input = SortedDict(dic)
    return tuple(function_input.values())


def make_namedtuple(class_name: str, **fields) -> namedtuple:
    """Convenience method for making a named tuple

    Args:
        class_name (str): name of the named tuple

    Returns:
        namedtuple: a named tuple with the fields as values
    """
    return namedtuple(class_name, fields)(*fields.values())


def get_nearest_coords(dataset: xr.Dataset, collapse_list=False, **kwargs) -> dict:
    """Given an xarray dataset and kwargs of key value pairs of coordinate values, return a dictionary of the nearest coordinate name value pair that was found in the dataset

    Args:
        ds (xr.Dataset): dataset

    Returns:
        dict: nearest coordinate name value pair that matches the input coordinate name value pairs.
    """

    selection = dataset.sel(method="nearest", **kwargs)
    cd = selection.coords.to_dataset().to_dict()["coords"]
    cd2 = {}
    for k, v in cd.items():
        cd2[k] = v["data"]
        if collapse_list and isinstance(cd2[k], list):
            cd2[k] = cd2[k][0]  # select the first item in the list
    return cd2


def get_nearest_coords1D(val: Any, coords) -> Any:
    if isinstance(val, (int, float)):
        return min(coords, key=lambda x_: abs(x_ - val))
    for i in coords:
        if val == i:
            return i
    return val


def hash_sha1(var: any) -> str:
    """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
    return hashlib.sha1(str(var).encode("ASCII")).hexdigest()


def capitalise_words(message: str):
    """Given a string of lowercase words, capitalise them

    Args:
        message (str): lower case string

    Returns:
        _type_: capitalised string
    """
    capitalized_message = " ".join([word.capitalize() for word in message.split(" ")])
    return capitalized_message


def un_camel(camel: str) -> str:
    """Given a snake_case string return a CamelCase string

    Args:
        camel (str): camelcase string

    Returns:
        str: uncamelcased string
    """

    return capitalise_words(re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", camel.replace("_", " ")))


def mult_tuple(inp: Tuple[float], val: float) -> Tuple[float]:
    return tuple(np.array(inp) * val)


def tabs_in_markdown(regular_str: str, spaces: int = 2) -> str:
    """Given a string with tabs in the form \t convert the to &ensp; which is a double space in markdown

    Args:
        regular_str (str): A string with tabs in it
        spaces (int): the number of spaces per tab

    Returns:
        str: A string with sets of &nbsp; to represent the tabs in markdown
    """
    return regular_str.replace("\t", "".join(["&nbsp;"] * spaces))


def int_to_col(int_val, sat=0.5, val=0.95, alpha=-1) -> tuple[float, float, float]:
    """Uses the golden angle to generate colors programmatically with minimum overlap between colors.
    https://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/

    Args:
        int_val (_type_): index of an object you want to color, this is mapped to hue in HSV
        sat (float, optional): saturation in HSV. Defaults to 0.5.
        val (float, optional): value in HSV. Defaults to 0.95.
        alpha (int, optional): transparency.  If -1 then only RGB is returned, if 0 or greater, RGBA is returned. Defaults to -1.

    Returns:
        tuple[float, float, float] | tuple[float, float, float, float]: either RGB or RGBA vector
    """
    golden_ratio_conjugate = (1 + math.sqrt(5)) / 2
    rgb = hsv_to_rgb(int_val * golden_ratio_conjugate, sat, val)
    if alpha >= 0:
        return (*rgb, alpha)
    return rgb


def lerp(value, input_low: float, input_high: float, output_low: float, output_high: float):
    input_low = float(input_low)
    return output_low + ((float(value) - input_low) / (float(input_high) - input_low)) * (
        float(output_high) - output_low
    )


def color_tuple_to_css(color: tuple[float, float, float]) -> str:
    return f"rgb{(color[0] * 255, color[1] * 255, color[2] * 255)}"


def color_tuple_to_255(color: tuple[float, float, float]) -> tuple[float, float, float]:
    return (
        min(int(color[0] * 255), 255),
        min(int(color[1] * 255), 255),
        min(int(color[2] * 255), 255),
    )


def gen_path(filename, folder="generic", suffix=".dat"):
    path = Path(f"cachedir/{folder}/{filename}/")
    path.mkdir(parents=True, exist_ok=True)
    return f"{path.absolute().as_posix()}/{filename}_{uuid4()}{suffix}"


def gen_video_path(video_name: str = "vid", extension: str = ".mp4") -> str:
    return gen_path(video_name, "vid", extension)


def gen_image_path(image_name: str = "img", filetype=".png") -> str:
    return gen_path(image_name, "img", filetype)


def gen_rerun_data_path(rrd_name: str = "rrd", filetype=".rrd") -> str:
    return gen_path(rrd_name, "rrd", filetype)


def callable_name(any_callable: Callable[..., Any]) -> str:
    if isinstance(any_callable, partial):
        return any_callable.func.__name__
    try:
        return any_callable.__name__
    except AttributeError:
        return str(any_callable)


def listify(obj) -> list:
    """Take an object and turn it into a list if its not already a list.  However if the object is none, don't turn it into a list"""
    if obj is None:
        return None
    if isinstance(obj, list):
        return obj
    if isinstance(obj, tuple):
        return list(obj)
    return [obj]


def get_name(var):
    if isinstance(var, param.Parameter):
        return var.name
    return var


def params_to_str(param_list: List[param.Parameter]):
    return [get_name(i) for i in param_list]


def publish_file(filepath: str, remote: str, branch_name: str) -> str:  # pragma: no cover
    """Publish a file to an orphan git branch:

    .. code-block:: python

        def publish_args(branch_name) -> Tuple[str, str]:
            return (
                "https://github.com/dyson-ai/bencher.git",
                f"https://github.com/dyson-ai/bencher/blob/{branch_name}")


    Args:
        remote (Callable): A function the returns a tuple of the publishing urls. It must follow the signature def publish_args(branch_name) -> Tuple[str, str].  The first url is the git repo name, the second url needs to match the format for viewable html pages on your git provider.  The second url can use the argument branch_name to point to the file on a specified branch.

    Returns:
        str: the url of the published file
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        shutil.copy(filepath, temp_dir)
        filename = Path(filepath).name
        filepath_tmp = Path(temp_dir) / filename

        logging.info(f"created report at: {filepath_tmp.absolute()}")
        cd_dir = f"cd {temp_dir} &&"

        # create a new git repo and add files to that.  Push the file to another arbitrary repo.  The aim of doing it this way is that no data needs to be downloaded.

        # os.system(f"{cd_dir} git config init.defaultBranch {branch_name}")
        os.system(f"{cd_dir} git init")
        os.system(f"{cd_dir} git branch -m {branch_name}")
        os.system(f"{cd_dir} git add {filename}")
        os.system(f'{cd_dir} git commit -m "publish {branch_name}"')
        os.system(f"{cd_dir} git remote add origin {remote}")
        os.system(f"{cd_dir} git push --set-upstream origin {branch_name} -f")


def github_content(remote: str, branch_name: str, filename: str):  # pragma: no cover
    raw = remote.replace(".git", "").replace(
        "https://github.com/", "https://raw.githubusercontent.com/"
    )
    return f"{raw}/{branch_name}/{filename}?token=$(date +%s)"


# import logging
# # from rerun.legacy_notebook import as_html
# import rerun as rr
# import panel as pn
# # from .utils import publish_file, gen_rerun_data_path


# def rrd_to_pane(
#     url: str, width: int = 499, height: int = 600, version: str = None
# ):  # pragma: no cover
#     if version is None:
#         version = "-1.20.1"  # TODO find a better way of doing this
#     return pn.pane.HTML(
#         f'<iframe src="https://app.rerun.io/version/{version}/?url={url}" width={width} height={height}></iframe>'
#     )


# # def to_pane(path: str):
# #     as_html()
# #     return rrd_to_pane(path)


# def publish_and_view_rrd(
#     file_path: str,
#     remote: str,
#     branch_name,
#     content_callback: callable,
#     version: str = None,
# ):  # pragma: no cover
#     as_html()
#     publish_file(file_path, remote=remote, branch_name="test_rrd")
#     publish_path = content_callback(remote, branch_name, file_path)
#     logging.info(publish_path)
#     return rrd_to_pane(publish_path, version=version)


# def record_rerun_session():
#     rrd_path = gen_rerun_data_path()
#     rr.save(rrd_path)
#     path = rrd_path.split("cachedir")[0]
#     return rrd_to_pane(f"http://126.0.0.1:8001/{path}")



import logging
from rerun.legacy_notebook import as_html
import rerun as rr
import panel as pn
from .utils import publish_file, gen_rerun_data_path


def rrd_to_pane(
    url: str, width: int = 500, height: int = 600, version: str = None
):  # pragma: no cover
    if version is None:
        version = "0.20.1"  # TODO find a better way of doing this
    return pn.pane.HTML(
        f'<iframe src="https://app.rerun.io/version/{version}/?url={url}" width={width} height={height}></iframe>'
    )


def to_pane(path: str):
    as_html()
    return rrd_to_pane(path)


def publish_and_view_rrd(
    file_path: str,
    remote: str,
    branch_name,
    content_callback: callable,
    version: str = None,
):  # pragma: no cover
    as_html()
    publish_file(file_path, remote=remote, branch_name="test_rrd")
    publish_path = content_callback(remote, branch_name, file_path)
    logging.info(publish_path)
    return rrd_to_pane(publish_path, version=version)


def capture_rerun_window(width: int = 500, height: int = 500):
    rrd_path = gen_rerun_data_path()
    rr.save(rrd_path)
    path = rrd_path.split("cachedir")[1]
    return rrd_to_pane(f"http://127.0.0.1:8001/{path}", width=width, height=height)



import rerun as rr
import bencher as bch
import panel as pn


rr.init("rerun_example_local", spawn=True)
file_path = "dat1.rrd"
rr.save(file_path)

rr.log("s1", rr.Scalar(1))
rr.log("s1", rr.Scalar(4))
rr.log("s1", rr.Scalar(2))

local = True

if local:
    row = pn.Row()
    # row.append(rrd_to_pane("http://localhost:8001/dat2.rrd"))

    row.append(bch.rrd_to_pane("http://127.0.0.1:8001/dat2.rrd"))
    row.show()
else:
    # publish data to a github branch
    bch.publish_and_view_rrd(
        file_path,
        remote="https://github.com/dyson-ai/bencher.git",
        branch_name="test_rrd",
        content_callback=bch.github_content,
    ).show()



from typing import Optional, List
from copy import deepcopy
import panel as pn
import xarray as xr
from param import Parameter
from bencher.results.bench_result_base import BenchResultBase, ReduceType
from bencher.variables.results import ResultImage
from bencher.plotting.plot_filter import VarRange, PlotFilter
from bencher.utils import callable_name, int_to_col, color_tuple_to_255
from bencher.video_writer import VideoWriter
from bencher.results.video_result import VideoControls
from bencher.results.composable_container.composable_container_video import (
    ComposableContainerVideo,
    ComposeType,
    RenderCfg,
)


class VideoSummaryResult(BenchResultBase):
    def to_video_summary(
        self,
        result_var: Parameter = None,
        reverse: bool = True,
        result_types=(ResultImage,),
        **kwargs,
    ) -> Optional[pn.panel]:
        return self.to_video_grid(
            result_var=result_var,
            result_types=result_types,
            time_sequence_dimension=-1,
            reverse=reverse,
            **kwargs,
        )

    def to_video_grid(
        self,
        result_var: Parameter = None,
        result_types=(ResultImage,),
        pane_collection: pn.pane = None,
        time_sequence_dimension=0,
        target_duration: float = None,
        **kwargs,
    ) -> Optional[pn.panel]:
        """Returns the results compiled into a video

        Args:
            result_var (Parameter, optional): The result var to plot. Defaults to None.
            result_types (tuple, optional): The types of result var to convert to video. Defaults to (ResultImage,).
            collection (pn.pane, optional): If there are multiple results, use this collection to stack them. Defaults to pn.Row().

        Returns:
            Optional[pn.panel]: a panel pane with a video of all results concatenated together
        """
        plot_filter = PlotFilter(
            float_range=VarRange(0, None),
            cat_range=VarRange(0, None),
            panel_range=VarRange(1, None),
            input_range=VarRange(1, None),
        )
        matches_res = plot_filter.matches_result(
            self.plt_cnt_cfg, callable_name(self.to_video_grid_ds)
        )

        if pane_collection is None:
            pane_collection = pn.Row()

        if matches_res.overall:
            ds = self.to_dataset(ReduceType.SQUEEZE)
            for rv in self.get_results_var_list(result_var):
                if isinstance(rv, result_types):
                    pane_collection.append(
                        self.to_video_grid_ds(
                            ds,
                            rv,
                            time_sequence_dimension=time_sequence_dimension,
                            target_duration=target_duration,
                            **kwargs,
                        )
                    )
            return pane_collection
        return matches_res.to_panel()

    def to_video_grid_ds(
        self,
        dataset: xr.Dataset,
        result_var: Parameter,
        reverse=True,
        time_sequence_dimension=0,
        video_controls: VideoControls = None,
        target_duration: float = None,
        **kwargs,
    ):
        cvc = self._to_video_panes_ds(
            dataset,
            self.plot_cb,
            target_dimension=0,
            horizontal=True,
            compose_method=ComposeType.right,
            time_sequence_dimension=time_sequence_dimension,
            result_var=result_var,
            final=True,
            reverse=reverse,
            target_duration=target_duration,
            **kwargs,
        )

        filename = VideoWriter().write_video_raw(cvc)

        if filename is not None:
            if video_controls is None:
                video_controls = VideoControls()
            return video_controls.video_container(
                filename, width=kwargs.get("width", None), height=kwargs.get("height", None)
            )
        return None

    def plot_cb(self, dataset, result_var, **kwargs):
        val = self.ds_to_container(dataset, result_var, container=None, **kwargs)
        return val

    def dataset_to_compose_list(
        self,
        dataset: xr.Dataset,
        first_compose_method: ComposeType = ComposeType.down,
        time_sequence_dimension: int = 0,
    ) -> List[ComposeType]:
        """ "Given a dataset, chose an order for composing the results.  By default will flip between right and down and the last dimension will be a time sequence.

        Args:
            dataset (xr.Dataset): the dataset to render
            first_compose_method (ComposeType, optional): the direction of the first composition method. Defaults to ComposeType.right.
            time_sequence_dimension (int, optional): The dimension to start time sequencing instead of composing in space. Defaults to 0.

        Returns:
            List[ComposeType]: A list of composition methods for composing the dataset result
        """

        num_dims = len(dataset.sizes)
        if time_sequence_dimension == -1:  # use time sequence for everything
            compose_method_list = [ComposeType.sequence] * (num_dims + 1)
        else:
            compose_method_list = [first_compose_method]
            compose_method_list.extend(
                ComposeType.flip(compose_method_list[-1]) for _ in range(num_dims - 1)
            )
            compose_method_list.append(ComposeType.sequence)

            for i in range(min(len(compose_method_list), time_sequence_dimension + 1)):
                compose_method_list[i] = ComposeType.sequence

        return compose_method_list

    def _to_video_panes_ds(
        self,
        dataset: xr.Dataset,
        plot_callback: callable = None,
        target_dimension=0,
        compose_method=ComposeType.right,
        compose_method_list=None,
        result_var=None,
        time_sequence_dimension=0,
        root_dimensions=None,
        reverse=False,
        target_duration: float = None,
        **kwargs,
    ) -> pn.panel:
        num_dims = len(dataset.sizes)
        dims = list(d for d in dataset.sizes)
        if reverse:
            dims = list(reversed(dims))

        if root_dimensions is None:
            root_dimensions = num_dims

        if compose_method_list is None:
            compose_method_list = self.dataset_to_compose_list(
                dataset, compose_method, time_sequence_dimension=time_sequence_dimension
            )

            # print(compose_method_list)

        compose_method_list_pop = deepcopy(compose_method_list)
        if len(compose_method_list_pop) > 1:
            compose_method = compose_method_list_pop.pop()

        if num_dims > (target_dimension) and num_dims != 0:
            selected_dim = dims[-1]
            outer_container = ComposableContainerVideo()
            for i in range(dataset.sizes[selected_dim]):
                sliced = dataset.isel({selected_dim: i})
                label_val = sliced.coords[selected_dim].values.item()
                inner_container = ComposableContainerVideo()

                panes = self._to_video_panes_ds(
                    sliced,
                    plot_callback=plot_callback,
                    target_dimension=target_dimension,
                    compose_method_list=compose_method_list_pop,
                    result_var=result_var,
                    root_dimensions=root_dimensions,
                    time_sequence_dimension=time_sequence_dimension,
                )
                inner_container.append(panes)

                rendered = inner_container.render(
                    RenderCfg(
                        var_name=selected_dim,
                        var_value=label_val,
                        compose_method=compose_method,
                        duration=target_duration,
                    )
                )
                outer_container.append(rendered)
            return outer_container.render(
                RenderCfg(
                    compose_method=compose_method,
                    duration=target_duration,
                    background_col=color_tuple_to_255(int_to_col(num_dims - 2, 0.05, 1.0)),
                    # background_col= (255,0,0),
                )
            )
        return plot_callback(dataset=dataset, result_var=result_var, **kwargs)



from flask import Flask, send_from_directory
from flask_cors import CORS
import threading
from pathlib import Path


def create_server(directory):
    """Create a Flask app configured to serve files from the specified directory."""
    app = Flask(__name__)
    CORS(app)

    @app.route("/<path:path>", methods=["GET"])
    def serve_file(path):
        return send_from_directory(directory, path)

    return app


def run_flask_in_thread(directory=None, port=8001):
    """Run the Flask server to serve files from the specified directory."""
    if directory is None:
        directory = Path("cachedir/").absolute().as_posix()
    app = create_server(directory)
    threading.Thread(target=app.run, kwargs={"port": port}, daemon=True).start()
    print(f"Flask server is running on port {port} serving files from {directory}")



from typing import List, Tuple, Any
from dataclasses import dataclass, field
from sortedcontainers import SortedDict
from .utils import hash_sha1
from bencher.utils import hmap_canonical_input


@dataclass
class WorkerJob:
    function_input_vars: List
    index_tuple: Tuple[int]
    dims_name: List[str]
    constant_inputs: dict
    bench_cfg_sample_hash: str
    tag: str

    function_input: SortedDict = None
    canonical_input: Tuple[Any] = None
    fn_inputs_sorted: List[str] = None
    function_input_signature_pure: str = None
    function_input_signature_benchmark_context: str = None
    found_in_cache: bool = False
    msgs: List[str] = field(default_factory=list)

    def setup_hashes(self) -> None:
        self.function_input = SortedDict(zip(self.dims_name, self.function_input_vars))

        self.canonical_input = hmap_canonical_input(self.function_input)

        if self.constant_inputs is not None:
            self.function_input = self.function_input | self.constant_inputs

        # store a tuple of the inputs as keys for a holomap
        # the signature is the hash of the inputs to to the function + meta variables such as repeat and time + the hash of the benchmark sweep as a whole (without the repeats hash)
        self.fn_inputs_sorted = list(SortedDict(self.function_input).items())
        self.function_input_signature_pure = hash_sha1((self.fn_inputs_sorted, self.tag))

        self.function_input_signature_benchmark_context = hash_sha1(
            (self.function_input_signature_pure, self.bench_cfg_sample_hash)
        )



from diskcache import Cache

from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.utils import hash_sha1
import logging
from sortedcontainers import SortedDict

# from job import job,JobCache,JobFunctionCache


class CachedParams(ParametrizedSweep):
    def __init__(self, clear_cache=True, cache_name="fcache", **params):
        super().__init__(**params)

        self.cache = Cache(f"cachedir/{cache_name}/sample_cache")
        logging.info(f"cache dir{self.cache.directory}")
        print(f"cache dir{self.cache.directory}")

        if clear_cache:
            self.cache.clear()

    def kwargs_to_hash_key(self, **kwargs):
        return tuple(SortedDict(kwargs).items())

    def in_cache(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)
        # print(self.get_inputs_as_dict())
        inputs_key = self.kwargs_to_hash_key(**self.get_inputs_as_dict())
        key = hash_sha1(inputs_key)
        print(f"key:{key},value: {inputs_key}")
        value = None
        if key in self.cache:
            value = self.cache[key]
        return key, value

    def cache_wrap(self, func, **kwargs):
        key, value = self.in_cache(**kwargs)
        if value is None:
            value = func(**kwargs)
            self.cache[key] = value
        return value

    # def cache_mem(self, function):
    #     def cache_wrap1(self, func, **kwargs):
    #         key, value = self.in_cache(**kwargs)
    #         if value is None:
    #             value = function(**kwargs)
    #             self.cache[key] = value
    #         return value

    #     return cache_wrap1



import logging
from datetime import datetime
from itertools import product, combinations

from typing import Callable, List, Optional
from copy import deepcopy
import numpy as np
import param
import xarray as xr
from diskcache import Cache
from contextlib import suppress
from functools import partial
import panel as pn

from bencher.worker_job import WorkerJob

from bencher.bench_cfg import BenchCfg, BenchRunCfg, DimsCfg
from bencher.bench_plot_server import BenchPlotServer
from bencher.bench_report import BenchReport

from bencher.variables.inputs import IntSweep
from bencher.variables.time import TimeSnapshot, TimeEvent
from bencher.variables.results import (
    ResultVar,
    ResultVec,
    ResultHmap,
    ResultPath,
    ResultVideo,
    ResultImage,
    ResultString,
    ResultContainer,
    ResultReference,
    ResultDataSet,
)
from bencher.results.bench_result import BenchResult
from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.job import Job, FutureCache, JobFuture, Executors
from bencher.utils import params_to_str

# Customize the formatter
formatter = logging.Formatter("%(levelname)s: %(message)s")
logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")


for handler in logging.root.handlers:
    handler.setFormatter(formatter)


def set_xarray_multidim(data_array: xr.DataArray, index_tuple, value: float) -> xr.DataArray:
    # """This is terrible, I need to do this in a better way, but [] does not like *args syntax and the () version of the set function doesn't either"""
    match len(index_tuple):
        case 1:
            data_array[index_tuple[0]] = value
        case 2:
            data_array[index_tuple[0], index_tuple[1]] = value
        case 3:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2]] = value
        case 4:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3]] = value
        case 5:
            data_array[
                index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3], index_tuple[4]
            ] = value
        case 6:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
            ] = value
        case 7:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
            ] = value
        case 8:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
            ] = value
        case 9:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
                index_tuple[8],
            ] = value

    return data_array


def kwargs_to_input_cfg(worker_input_cfg: ParametrizedSweep, **kwargs) -> ParametrizedSweep:
    input_cfg = worker_input_cfg()
    input_cfg.param.update(kwargs)
    return input_cfg


def worker_cfg_wrapper(worker, worker_input_cfg: ParametrizedSweep, **kwargs) -> dict:
    input_cfg = kwargs_to_input_cfg(worker_input_cfg, **kwargs)
    return worker(input_cfg)


def worker_kwargs_wrapper(worker: Callable, bench_cfg: BenchCfg, **kwargs) -> dict:
    function_input_deep = deepcopy(kwargs)
    if not bench_cfg.pass_repeat:
        function_input_deep.pop("repeat")
    if "over_time" in function_input_deep:
        function_input_deep.pop("over_time")
    if "time_event" in function_input_deep:
        function_input_deep.pop("time_event")
    return worker(**function_input_deep)


class Bench(BenchPlotServer):
    def __init__(
        self,
        bench_name: str = None,
        worker: Callable | ParametrizedSweep = None,
        worker_input_cfg: ParametrizedSweep = None,
        run_cfg=None,
        report=None,
    ) -> None:
        """Create a new Bench object from a function and a class defining the inputs to the function

        Args:
            bench_name (str): The name of the benchmark and output folder for the figures
            worker (Callable | ParametrizedSweep): A function that accepts a class of type (worker_input_config)
            worker_input_config (ParametrizedSweep): A class defining the parameters of the function.
        """
        assert isinstance(bench_name, str)
        self.bench_name = bench_name
        self.worker = None
        self.worker_class_instance = None
        self.worker_input_cfg = None
        self.worker_class_instance = None
        self.set_worker(worker, worker_input_cfg)
        self.run_cfg = run_cfg
        if report is None:
            self.report = BenchReport(self.bench_name)
        else:
            self.report = report
            if self.report.bench_name is None:
                self.report.bench_name = self.bench_name
        self.results = []

        self.bench_cfg_hashes = []  # a list of hashes that point to benchmark results
        self.last_run_cfg = None  # cached run_cfg used to pass to the plotting function
        self.sample_cache = None  # store the results of each benchmark function call in a cache
        self.ds_dynamic = {}  # A dictionary to store unstructured vector datasets

        self.cache_size = int(100e9)  # default to 100gb

        # self.bench_cfg = BenchCfg()

        # Maybe put this in SweepCfg
        self.input_vars = None
        self.result_vars = None
        self.const_vars = None
        self.plot_callbacks = []
        self.plot = True

    def add_plot_callback(self, callback: Callable[[BenchResult], pn.panel], **kwargs) -> None:
        """Add a plotting callback that will be called on any result produced when calling a sweep function.  You can pass additional arguments to the plotting function with kwargs.  e.g.  add_plot_callback(bch.BenchResult.to_video_grid,)

        Args:
            callback (Callable[[BenchResult], pn.panel]): _description_
        """
        self.plot_callbacks.append(partial(callback, **kwargs))

    def set_worker(self, worker: Callable, worker_input_cfg: ParametrizedSweep = None) -> None:
        """Set the benchmark worker function and optionally the type the worker expects

        Args:
            worker (Callable): The benchmark worker function
            worker_input_cfg (ParametrizedSweep, optional): The input type the worker expects. Defaults to None.
        """

        if isinstance(worker, ParametrizedSweep):
            self.worker_class_instance = worker
            # self.worker_class_type = type(worker)
            self.worker = self.worker_class_instance.__call__
            logging.info("setting worker from bench class.__call__")
        else:
            if isinstance(worker, type):
                raise RuntimeError("This should be a class instance, not a class")
            if worker_input_cfg is None:
                self.worker = worker
            else:
                self.worker = partial(worker_cfg_wrapper, worker, worker_input_cfg)
            logging.info(f"setting worker {worker}")
        self.worker_input_cfg = worker_input_cfg

    def sweep_sequential(
        self,
        title="",
        input_vars: List[ParametrizedSweep] = None,
        result_vars: List[ParametrizedSweep] = None,
        const_vars: List[ParametrizedSweep] = None,
        optimise_var: ParametrizedSweep = None,
        run_cfg: BenchRunCfg = None,
        group_size: int = 1,
        iterations: int = 1,
        relationship_cb=None,
        plot_callbacks: List | bool = None,
    ) -> List[BenchResult]:
        if relationship_cb is None:
            relationship_cb = combinations
        if input_vars is None:
            input_vars = self.worker_class_instance.get_inputs_only()
        results = []
        for it in range(iterations):
            for input_group in relationship_cb(input_vars, group_size):
                title_gen = title + "Sweeping " + " vs ".join(params_to_str(input_group))
                if iterations > 1:
                    title_gen += f" iteration:{it}"
                res = self.plot_sweep(
                    title=title_gen,
                    input_vars=list(input_group),
                    result_vars=result_vars,
                    const_vars=const_vars,
                    run_cfg=run_cfg,
                    plot_callbacks=plot_callbacks,
                )

                if optimise_var is not None:
                    const_vars = res.get_optimal_inputs(optimise_var, True)
                results.append(res)
        return results

    def plot_sweep(
        self,
        title: str = None,
        input_vars: List[ParametrizedSweep] = None,
        result_vars: List[ParametrizedSweep] = None,
        const_vars: List[ParametrizedSweep] = None,
        time_src: datetime = None,
        description: str = None,
        post_description: str = None,
        pass_repeat: bool = False,
        tag: str = "",
        run_cfg: BenchRunCfg = None,
        plot_callbacks: List | bool = None,
    ) -> BenchResult:
        """The all in 1 function benchmarker and results plotter.

        Args:
            input_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            result_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            const_vars (List[ParametrizedSweep], optional): A list of variables to keep constant with a specified value. Defaults to None.
            title (str, optional): The title of the benchmark. Defaults to None.
            description (str, optional): A description of the benchmark. Defaults to None.
            post_description (str, optional): A description that comes after the benchmark plots. Defaults to None.
            time_src (datetime, optional): Set a time that the result was generated. Defaults to datetime.now().
            pass_repeat (bool,optional) By default do not pass the kwarg 'repeat' to the benchmark function.  Set to true if
            you want the benchmark function to be passed the repeat number
            tag (str,optional): Use tags to group different benchmarks together.
            run_cfg: (BenchRunCfg, optional): A config for storing how the benchmarks and run
            plot_callbacks: (List | bool) A list of plot callbacks to call on the results. Pass false or an empty list to turn off plotting
        Raises:
            ValueError: If a result variable is not set

        Returns:
            BenchResult: A class with all the data used to generate the results and the results
        """

        input_vars_in = deepcopy(input_vars)
        result_vars_in = deepcopy(result_vars)
        const_vars_in = deepcopy(const_vars)

        if self.worker_class_instance is not None:
            if input_vars_in is None:
                logging.info(
                    "No input variables passed, using all param variables in bench class as inputs"
                )
                if self.input_vars is None:
                    input_vars_in = self.worker_class_instance.get_inputs_only()
                else:
                    input_vars_in = deepcopy(self.input_vars)
                for i in input_vars_in:
                    logging.info(f"input var: {i.name}")
            if result_vars_in is None:
                logging.info(
                    "No results variables passed, using all result variables in bench class:"
                )
                if self.result_vars is None:
                    result_vars_in = self.worker_class_instance.get_results_only()
                else:
                    result_vars_in = deepcopy(self.result_vars)

            if const_vars_in is None:
                if self.const_vars is None:
                    const_vars_in = self.worker_class_instance.get_input_defaults()
                else:
                    const_vars_in = deepcopy(self.const_vars)
        else:
            if input_vars_in is None:
                input_vars_in = []
            if result_vars_in is None:
                result_vars_in = []
            if const_vars_in is None:
                const_vars_in = []

        if run_cfg is None:
            if self.run_cfg is None:
                run_cfg = BenchRunCfg()
                logging.info("Generate default run cfg")
            else:
                run_cfg = deepcopy(self.run_cfg)
                logging.info("Copy run cfg from bench class")

        if run_cfg.only_plot:
            run_cfg.use_cache = True

        self.last_run_cfg = run_cfg

        if isinstance(input_vars_in, dict):
            input_lists = []
            for k, v in input_vars_in.items():
                param_var = self.convert_vars_to_params(k, "input", run_cfg)
                if isinstance(v, list):
                    assert len(v) > 0
                    param_var = param_var.with_sample_values(v)

                else:
                    raise RuntimeError("Unsupported type")
                input_lists.append(param_var)

            input_vars_in = input_lists
        else:
            for i in range(len(input_vars_in)):
                input_vars_in[i] = self.convert_vars_to_params(input_vars_in[i], "input", run_cfg)
        for i in range(len(result_vars_in)):
            result_vars_in[i] = self.convert_vars_to_params(result_vars_in[i], "result", run_cfg)

        for r in result_vars_in:
            logging.info(f"result var: {r.name}")

        if isinstance(const_vars_in, dict):
            const_vars_in = list(const_vars_in.items())

        for i in range(len(const_vars_in)):
            # consts come as tuple pairs
            cv_list = list(const_vars_in[i])
            cv_list[0] = self.convert_vars_to_params(cv_list[0], "const", run_cfg)
            const_vars_in[i] = cv_list

        if title is None:
            if len(input_vars_in) > 0:
                title = "Sweeping " + " vs ".join([i.name for i in input_vars_in])
            elif len(const_vars_in) > 0:
                title = "Constant Value"
                if len(const_vars_in) > 1:
                    title += "s"
                title += ": " + ", ".join([f"{c[0].name}={c[1]}" for c in const_vars_in])
            else:
                raise RuntimeError("you must pass a title, or define inputs or consts")

        if run_cfg.level > 0:
            inputs = []
            print(input_vars_in)
            if len(input_vars_in) > 0:
                for i in input_vars_in:
                    inputs.append(i.with_level(run_cfg.level))
                input_vars_in = inputs

        # if any of the inputs have been include as constants, remove those variables from the list of constants
        with suppress(ValueError, AttributeError):
            for i in input_vars_in:
                for c in const_vars_in:
                    # print(i.hash_persistent())
                    if i.name == c[0].name:
                        const_vars_in.remove(c)
                        logging.info(f"removing {i.name} from constants")

        result_hmaps = []
        result_vars_only = []
        for i in result_vars_in:
            if isinstance(i, ResultHmap):
                result_hmaps.append(i)
            else:
                result_vars_only.append(i)

        if post_description is None:
            post_description = (
                "## Results Description\nPlease set post_description to explain these results"
            )

        if plot_callbacks is None:
            if self.plot_callbacks is not None and len(self.plot_callbacks) == 0:
                plot_callbacks = [BenchResult.to_auto_plots]
            else:
                plot_callbacks = self.plot_callbacks
        elif isinstance(plot_callbacks, bool):
            plot_callbacks = [BenchResult.to_auto_plots] if plot_callbacks else []

        bench_cfg = BenchCfg(
            input_vars=input_vars_in,
            result_vars=result_vars_only,
            result_hmaps=result_hmaps,
            const_vars=const_vars_in,
            bench_name=self.bench_name,
            description=description,
            post_description=post_description,
            title=title,
            pass_repeat=pass_repeat,
            tag=run_cfg.run_tag + tag,
            plot_callbacks=plot_callbacks,
        )
        return self.run_sweep(bench_cfg, run_cfg, time_src)

    def run_sweep(
        self, bench_cfg: BenchCfg, run_cfg: BenchRunCfg, time_src: datetime
    ) -> BenchResult:
        print("tag", bench_cfg.tag)

        bench_cfg.param.update(run_cfg.param.values())
        bench_cfg_hash = bench_cfg.hash_persistent(True)
        bench_cfg.hash_value = bench_cfg_hash

        # does not include repeats in hash as sample_hash already includes repeat as part of the per sample hash
        bench_cfg_sample_hash = bench_cfg.hash_persistent(False)

        if self.sample_cache is None:
            self.sample_cache = self.init_sample_cache(run_cfg)
        if bench_cfg.clear_sample_cache:
            self.clear_tag_from_sample_cache(bench_cfg.tag, run_cfg)

        calculate_results = True
        with Cache("cachedir/benchmark_inputs", size_limit=self.cache_size) as c:
            if run_cfg.clear_cache:
                c.delete(bench_cfg_hash)
                logging.info("cleared cache")
            elif run_cfg.use_cache:
                logging.info(
                    f"checking for previously calculated results with key: {bench_cfg_hash}"
                )
                if bench_cfg_hash in c:
                    logging.info(f"loading cached results from key: {bench_cfg_hash}")
                    bench_res = c[bench_cfg_hash]
                    # if not over_time:  # if over time we always want to calculate results
                    calculate_results = False
                else:
                    logging.info("did not detect results in cache")
                    if run_cfg.only_plot:
                        raise FileNotFoundError("Was not able to load the results to plot!")

        if calculate_results:
            if run_cfg.time_event is not None:
                time_src = run_cfg.time_event
            bench_res = self.calculate_benchmark_results(
                bench_cfg, time_src, bench_cfg_sample_hash, run_cfg
            )

            # use the hash of the inputs to look up historical values in the cache
            if run_cfg.over_time:
                bench_res.ds = self.load_history_cache(
                    bench_res.ds, bench_cfg_hash, run_cfg.clear_history
                )

            self.report_results(bench_res, run_cfg.print_xarray, run_cfg.print_pandas)
            self.cache_results(bench_res, bench_cfg_hash)

        logging.info(self.sample_cache.stats())
        self.sample_cache.close()

        bench_res.post_setup()

        if bench_cfg.auto_plot:
            self.report.append_result(bench_res)

        self.results.append(bench_res)
        return bench_res

    def convert_vars_to_params(
        self,
        variable: param.Parameter | str | dict | tuple,
        var_type: str,
        run_cfg: Optional[BenchRunCfg],
    ) -> param.Parameter:
        """check that a variable is a subclass of param

        Args:
            variable (param.Parameter): the variable to check
            var_type (str): a string representation of the variable type for better error messages

        Raises:
            TypeError: the input variable type is not a param.
        """
        if isinstance(variable, str):
            variable = self.worker_class_instance.param.objects(instance=False)[variable]
        if isinstance(variable, dict):
            param_var = self.worker_class_instance.param.objects(instance=False)[variable["name"]]
            if variable.get("values"):
                param_var = param_var.with_sample_values(variable["values"])

            if variable.get("samples"):
                param_var = param_var.with_samples(variable["samples"])
            if variable.get("max_level"):
                if run_cfg is not None:
                    param_var = param_var.with_level(run_cfg.level, variable["max_level"])
            variable = param_var
        if not isinstance(variable, param.Parameter):
            raise TypeError(
                f"You need to use {var_type}_vars =[{self.worker_input_cfg}.param.your_variable], instead of {var_type}_vars =[{self.worker_input_cfg}.your_variable]"
            )
        return variable

    def cache_results(self, bench_res: BenchResult, bench_cfg_hash: int) -> None:
        with Cache("cachedir/benchmark_inputs", size_limit=self.cache_size) as c:
            logging.info(f"saving results with key: {bench_cfg_hash}")
            self.bench_cfg_hashes.append(bench_cfg_hash)
            # object index may not be pickleable so remove before caching
            obj_index_tmp = bench_res.object_index
            bench_res.object_index = []

            c[bench_cfg_hash] = bench_res

            # restore object index
            bench_res.object_index = obj_index_tmp

            logging.info(f"saving benchmark: {self.bench_name}")
            c[self.bench_name] = self.bench_cfg_hashes

    # def show(self, run_cfg: BenchRunCfg = None, pane=None) -> None:
    #     """Launches a webserver with plots of the benchmark results, blocking

    #     Args:
    #         run_cfg (BenchRunCfg, optional): Options for the webserve such as the port. Defaults to None.

    #     """
    #     if run_cfg is None:
    #         if self.last_run_cfg is not None:
    #             run_cfg = self.last_run_cfg
    #         else:
    #             run_cfg = BenchRunCfg()

    #     return BenchPlotServer().plot_server(self.bench_name, run_cfg, pane)

    def load_history_cache(
        self, dataset: xr.Dataset, bench_cfg_hash: int, clear_history: bool
    ) -> xr.Dataset:
        """Load historical data from a cache if over_time=true

        Args:
            ds (xr.Dataset): Freshly calculated data
            bench_cfg_hash (int): Hash of the input variables used to generate the data
            clear_history (bool): Optionally clear the history

        Returns:
            xr.Dataset: historical data as an xr dataset
        """
        with Cache("cachedir/history", size_limit=self.cache_size) as c:
            if clear_history:
                logging.info("clearing history")
            else:
                logging.info(f"checking historical key: {bench_cfg_hash}")
                if bench_cfg_hash in c:
                    logging.info("loading historical data from cache")
                    ds_old = c[bench_cfg_hash]
                    dataset = xr.concat([ds_old, dataset], "over_time")
                else:
                    logging.info("did not detect any historical data")

            logging.info("saving data to history cache")
            c[bench_cfg_hash] = dataset
        return dataset

    def setup_dataset(
        self, bench_cfg: BenchCfg, time_src: datetime | str
    ) -> tuple[BenchResult, List, List]:
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime | str): a representation of the sample time

        Returns:
            tuple[BenchResult, List, List]: bench_result, function inputs, dimension names
        """

        if time_src is None:
            time_src = datetime.now()
        bench_cfg.meta_vars = self.define_extra_vars(bench_cfg, bench_cfg.repeats, time_src)

        bench_cfg.all_vars = bench_cfg.input_vars + bench_cfg.meta_vars
        # bench_cfg.all_vars = bench_cfg.iv_time + bench_cfg.input_vars +[ bench_cfg.iv_repeat]
        # bench_cfg.all_vars = [ bench_cfg.iv_repeat] +bench_cfg.input_vars + bench_cfg.iv_time

        for i in bench_cfg.all_vars:
            logging.info(i.sampling_str())

        dims_cfg = DimsCfg(bench_cfg)
        function_inputs = list(
            zip(product(*dims_cfg.dim_ranges_index), product(*dims_cfg.dim_ranges))
        )
        # xarray stores K N-dimensional arrays of data.  Each array is named and in this case we have an ND array for each result variable
        data_vars = {}
        dataset_list = []

        for rv in bench_cfg.result_vars:
            if isinstance(rv, ResultVar):
                result_data = np.full(dims_cfg.dims_size, np.nan, dtype=float)
                data_vars[rv.name] = (dims_cfg.dims_name, result_data)
            if isinstance(rv, (ResultReference, ResultDataSet)):
                result_data = np.full(dims_cfg.dims_size, -1, dtype=int)
                data_vars[rv.name] = (dims_cfg.dims_name, result_data)
            if isinstance(
                rv, (ResultPath, ResultVideo, ResultImage, ResultString, ResultContainer)
            ):
                result_data = np.full(dims_cfg.dims_size, "NAN", dtype=object)
                data_vars[rv.name] = (dims_cfg.dims_name, result_data)

            elif type(rv) is ResultVec:
                for i in range(rv.size):
                    result_data = np.full(dims_cfg.dims_size, np.nan)
                    data_vars[rv.index_name(i)] = (dims_cfg.dims_name, result_data)

        bench_res = BenchResult(bench_cfg)
        bench_res.ds = xr.Dataset(data_vars=data_vars, coords=dims_cfg.coords)
        bench_res.ds_dynamic = self.ds_dynamic
        bench_res.dataset_list = dataset_list
        bench_res.setup_object_index()

        return bench_res, function_inputs, dims_cfg.dims_name

    def define_const_inputs(self, const_vars) -> dict:
        constant_inputs = None
        if const_vars is not None:
            const_vars, constant_values = [
                [i for i, j in const_vars],
                [j for i, j in const_vars],
            ]

            constant_names = [i.name for i in const_vars]
            constant_inputs = dict(zip(constant_names, constant_values))
        return constant_inputs

    def define_extra_vars(self, bench_cfg: BenchCfg, repeats: int, time_src) -> list[IntSweep]:
        """Define extra meta vars that are stored in the n-d array but are not passed to the benchmarking function, such as number of repeats and the time the function was called.

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            repeats (int): the number of times to sample the function
            time_src (datetime): a representation of the sample time

        Returns:
            _type_: _description_
        """
        bench_cfg.iv_repeat = IntSweep(
            default=repeats,
            bounds=[1, repeats],
            samples=repeats,
            units="repeats",
            doc="The number of times a sample was measured",
        )
        bench_cfg.iv_repeat.name = "repeat"
        extra_vars = [bench_cfg.iv_repeat]

        if bench_cfg.over_time:
            if isinstance(time_src, str):
                iv_over_time = TimeEvent(time_src)
            else:
                iv_over_time = TimeSnapshot(time_src)
            iv_over_time.name = "over_time"
            extra_vars.append(iv_over_time)
            bench_cfg.iv_time = [iv_over_time]
        return extra_vars

    def calculate_benchmark_results(
        self, bench_cfg, time_src: datetime | str, bench_cfg_sample_hash, bench_run_cfg
    ) -> BenchResult:
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime): a representation of the sample time

        Returns:
            bench_cfg (BenchCfg): description of the benchmark parameters
        """
        bench_res, func_inputs, dims_name = self.setup_dataset(bench_cfg, time_src)
        bench_res.bench_cfg.hmap_kdims = sorted(dims_name)
        constant_inputs = self.define_const_inputs(bench_res.bench_cfg.const_vars)
        callcount = 1

        results_list = []
        jobs = []

        for idx_tuple, function_input_vars in func_inputs:
            job = WorkerJob(
                function_input_vars,
                idx_tuple,
                dims_name,
                constant_inputs,
                bench_cfg_sample_hash,
                bench_res.bench_cfg.tag,
            )
            job.setup_hashes()
            jobs.append(job)

            jid = f"{bench_res.bench_cfg.title}:call {callcount}/{len(func_inputs)}"
            worker = partial(worker_kwargs_wrapper, self.worker, bench_res.bench_cfg)
            cache_job = Job(
                job_id=jid,
                function=worker,
                job_args=job.function_input,
                job_key=job.function_input_signature_pure,
                tag=job.tag,
            )
            result = self.sample_cache.submit(cache_job)
            results_list.append(result)
            callcount += 1

            if bench_run_cfg.executor == Executors.SERIAL:
                self.store_results(result, bench_res, job, bench_run_cfg)

        if bench_run_cfg.executor != Executors.SERIAL:
            for job, res in zip(jobs, results_list):
                self.store_results(res, bench_res, job, bench_run_cfg)

        for inp in bench_res.bench_cfg.all_vars:
            self.add_metadata_to_dataset(bench_res, inp)

        return bench_res

    def store_results(
        self,
        job_result: JobFuture,
        bench_res: BenchResult,
        worker_job: WorkerJob,
        bench_run_cfg: BenchRunCfg,
    ) -> None:
        result = job_result.result()
        if result is not None:
            logging.info(f"{job_result.job.job_id}:")
            if bench_res.bench_cfg.print_bench_inputs:
                for k, v in worker_job.function_input.items():
                    logging.info(f"\t {k}:{v}")

            result_dict = result if isinstance(result, dict) else result.param.values()

            for rv in bench_res.bench_cfg.result_vars:
                result_value = result_dict[rv.name]
                if bench_run_cfg.print_bench_results:
                    logging.info(f"{rv.name}: {result_value}")

                if isinstance(
                    rv,
                    (
                        ResultVar,
                        ResultVideo,
                        ResultImage,
                        ResultString,
                        ResultContainer,
                        ResultPath,
                    ),
                ):
                    set_xarray_multidim(bench_res.ds[rv.name], worker_job.index_tuple, result_value)
                elif isinstance(rv, ResultDataSet):
                    bench_res.dataset_list.append(result_value)
                    set_xarray_multidim(
                        bench_res.ds[rv.name],
                        worker_job.index_tuple,
                        len(bench_res.dataset_list) - 1,
                    )
                elif isinstance(rv, ResultReference):
                    bench_res.object_index.append(result_value)
                    set_xarray_multidim(
                        bench_res.ds[rv.name],
                        worker_job.index_tuple,
                        len(bench_res.object_index) - 1,
                    )

                elif isinstance(rv, ResultVec):
                    if isinstance(result_value, (list, np.ndarray)):
                        if len(result_value) == rv.size:
                            for i in range(rv.size):
                                set_xarray_multidim(
                                    bench_res.ds[rv.index_name(i)],
                                    worker_job.index_tuple,
                                    result_value[i],
                                )

                else:
                    raise RuntimeError("Unsupported result type")
            for rv in bench_res.result_hmaps:
                bench_res.hmaps[rv.name][worker_job.canonical_input] = result_dict[rv.name]

            # bench_cfg.hmap = bench_cfg.hmaps[bench_cfg.result_hmaps[0].name]

    def init_sample_cache(self, run_cfg: BenchRunCfg):
        return FutureCache(
            overwrite=run_cfg.overwrite_sample_cache,
            executor=run_cfg.executor,
            cache_name="sample_cache",
            tag_index=True,
            size_limit=self.cache_size,
            use_cache=run_cfg.use_sample_cache,
        )

    def clear_tag_from_sample_cache(self, tag: str, run_cfg):
        """Clear all samples from the cache that match a tag
        Args:
            tag(str): clear samples with this tag
        """
        if self.sample_cache is None:
            self.sample_cache = self.init_sample_cache(run_cfg)
        self.sample_cache.clear_tag(tag)

    def add_metadata_to_dataset(self, bench_res: BenchResult, input_var: ParametrizedSweep) -> None:
        """Adds variable metadata to the xrarry so that it can be used to automatically plot the dimension units etc.

        Args:
            bench_cfg (BenchCfg):
            input_var (ParametrizedSweep): The variable to extract metadata from
        """

        for rv in bench_res.bench_cfg.result_vars:
            if type(rv) is ResultVar:
                bench_res.ds[rv.name].attrs["units"] = rv.units
                bench_res.ds[rv.name].attrs["long_name"] = rv.name
            elif type(rv) is ResultVec:
                for i in range(rv.size):
                    bench_res.ds[rv.index_name(i)].attrs["units"] = rv.units
                    bench_res.ds[rv.index_name(i)].attrs["long_name"] = rv.name
            else:
                pass  # todo

        dsvar = bench_res.ds[input_var.name]
        dsvar.attrs["long_name"] = input_var.name
        if input_var.units is not None:
            dsvar.attrs["units"] = input_var.units
        if input_var.__doc__ is not None:
            dsvar.attrs["description"] = input_var.__doc__

    def report_results(self, bench_cfg: BenchCfg, print_xarray: bool, print_pandas: bool):
        """Optionally display the calculated benchmark data as either as pandas, xarray or plot

        Args:
            bench_cfg (BenchCfg):
            print_xarray (bool):
            print_pandas (bool):
        """
        if print_xarray:
            logging.info(bench_cfg.ds)
        if print_pandas:
            logging.info(bench_cfg.ds.to_dataframe())

    def clear_call_counts(self) -> None:
        """Clear the worker and cache call counts, to help debug and assert caching is happening properly"""
        self.sample_cache.clear_call_counts()

    def get_result(self, index: int = -1) -> BenchResult:
        return self.results[index]

    def publish(self, remote_callback: Callable) -> str:
        branch_name = f"{self.bench_name}_{self.run_cfg.run_tag}"
        return self.report.publish(remote_callback, branch_name=branch_name)



import bencher as bch
import numpy as np
import math
import matplotlib.pyplot as plt


def polygon_points(radius: float, sides: int, start_angle: float):
    points = []
    for ang in np.linspace(0, 360, sides + 1):
        angle = math.radians(start_angle + ang)
        points.append(([math.sin(angle) * radius, math.cos(angle) * radius]))
    return points


class BenchPolygons(bch.ParametrizedSweep):
    sides = bch.IntSweep(default=3, bounds=(3, 7))
    radius = bch.FloatSweep(default=1, bounds=(0.2, 1))
    linewidth = bch.FloatSweep(default=1, bounds=(1, 10))
    linestyle = bch.StringSweep(["solid", "dashed", "dotted"])
    color = bch.StringSweep(["red", "green", "blue"])
    start_angle = bch.FloatSweep(default=0, bounds=[0, 360])
    polygon = bch.ResultImage()
    area = bch.ResultVar()
    side_length = bch.ResultVar()

    def __call__(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)
        points = polygon_points(self.radius, self.sides, self.start_angle)
        # self.hmap = hv.Curve(points)
        self.polygon = self.points_to_polygon_png(points, bch.gen_image_path("polygon"))

        self.side_length = 2 * self.radius * math.sin(math.pi / self.sides)
        self.area = (self.sides * self.side_length**2) / (4 * math.tan(math.pi / self.sides))
        return super().__call__()

    def points_to_polygon_png(self, points: list[float], filename: str):
        """Draw a closed polygon and save to png"""
        fig = plt.figure(frameon=False)
        ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0], frameon=False)
        ax.set_axis_off()
        ax.plot(
            [p[0] for p in points],
            [p[1] for p in points],
            linewidth=self.linewidth,
            linestyle=self.linestyle,
            color=self.color,
        )
        ax.set_xlim(-1, 1)
        ax.set_ylim(-1, 1)

        ax.set_aspect("equal")
        fig.add_axes(ax)
        fig.savefig(filename, dpi=30)

        return filename


def example_image(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:
    run_cfg.use_cache = False
    bench = bch.Bench("polygons", BenchPolygons(), run_cfg=run_cfg, report=report)

    bench.result_vars = ["polygon", "area"]

    bench.add_plot_callback(bch.BenchResult.to_sweep_summary)
    # bench.add_plot_callback(bch.BenchResult.to_auto, level=2)
    bench.add_plot_callback(bch.BenchResult.to_panes, level=3)
    # bench.add_plot_callback(bch.BenchResult.to_panes)

    sweep_vars = ["sides", "radius", "linewidth", "color"]

    # sweep_vars = ["sides", "radius" ]

    for i in range(1, len(sweep_vars)):
        s = sweep_vars[:i]
        bench.plot_sweep(
            f"Polygons Sweeping {len(s)} Parameters",
            input_vars=s,
        )
        bench.report.append(bench.get_result().to_panes())

    return bench


def example_image_vid(run_cfg: bch.BenchRunCfg = None, report: bch.BenchReport = None) -> bch.Bench:
    bench = BenchPolygons().to_bench(run_cfg, report)
    bench.add_plot_callback(bch.BenchResult.to_sweep_summary)
    bench.add_plot_callback(
        bch.BenchResult.to_video_grid,
        target_duration=0.06,
        compose_method_list=[
            bch.ComposeType.right,
            bch.ComposeType.right,
            bch.ComposeType.sequence,
        ],
    )
    # from functools import partial
    # bench.add_plot_callback(bch.BenchResult.to_video_summary)
    # bench.add_plot_callback(bch.BenchResult.to_video_grid, time_sequence_dimension=0)
    # bench.add_plot_callback(bch.BenchResult.to_video_grid)
    # bench.add_plot_callback(bch.BenchResult.to_video_grid, time_sequence_dimension=2)
    # bench.add_plot_callback(bch.BenchResult.to_video_grid, time_sequence_dimension=3)

    bench.plot_sweep(input_vars=["radius"])
    # res = bench.plot_sweep(input_vars=["radius"], plot=False)
    # bench.report.append(res.to_video_grid(target_duration=0.06))
    bench.plot_sweep(input_vars=["radius", "sides"])
    # bench.plot_sweep(input_vars=["radius", "sides", "linewidth"])
    # bench.plot_sweep(input_vars=["radius", "sides", "linewidth", "color"])

    return bench


if __name__ == "__main__":

    def simple():
        bench = BenchPolygons().to_bench(bch.BenchRunCfg(level=4))

        # bench.plot_sweep(input_vars=["sides","color","radius"])

        # res = bench.sweep(input_vars=["sides", "radius"])

        bench.plot_sweep(input_vars=["sides"])
        bench.plot_sweep(input_vars=["sides", "color"])

        res = bench.plot_sweep(input_vars=["sides", "radius"])
        bench.report.append(res.to_heatmap(target_dimension=3))
        bench.report.append(res.to_line(target_dimension=1))

        return bench

    def example_image_vid_sequential(
        run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
    ) -> bch.Bench:
        bench = BenchPolygons().to_bench(run_cfg, report)
        bench.add_plot_callback(bch.BenchResult.to_title)
        bench.add_plot_callback(bch.BenchResult.to_video_grid)
        bench.sweep_sequential(input_vars=["radius", "sides", "linewidth", "color"], group_size=4)
        return bench

    # def example_image_pairs()

    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.use_sample_cache = True
    # ex_run_cfg.debug = True
    # ex_run_cfg.repeats = 2
    ex_run_cfg.level = 4
    # example_image_vid(ex_run_cfg).report.show()
    simple().report.show()

    # example_image_vid_sequential(ex_run_cfg).report.show()
    # example_image(ex_run_cfg).report.show()



from __future__ import annotations

import argparse
import logging

from typing import List

import param
from str2bool import str2bool
import panel as pn


from bencher.variables.sweep_base import hash_sha1, describe_variable
from bencher.variables.time import TimeSnapshot, TimeEvent
from bencher.variables.results import OptDir
from bencher.job import Executors
from datetime import datetime

# from bencher.results.bench_result import BenchResult


class BenchPlotSrvCfg(param.Parameterized):
    port: int = param.Integer(None, doc="The port to launch panel with")
    allow_ws_origin = param.Boolean(
        False,
        doc="Add the port to the whilelist, (warning will disable remote access if set to true)",
    )
    show: bool = param.Boolean(True, doc="Open the served page in a web browser")


class BenchRunCfg(BenchPlotSrvCfg):
    """A Class to store options for how to run a benchmark parameter sweep"""

    repeats: bool = param.Integer(1, doc="The number of times to sample the inputs")

    over_time: bool = param.Boolean(
        False,
        doc="If true each time the function is called it will plot a timeseries of historical and the latest result.",
    )

    use_optuna: bool = param.Boolean(False, doc="show optuna plots")

    summarise_constant_inputs = param.Boolean(
        True, doc="Print the inputs that are kept constant when describing the sweep parameters"
    )

    print_bench_inputs: bool = param.Boolean(
        True, doc="Print the inputs to the benchmark function every time it is called"
    )

    print_bench_results: bool = param.Boolean(
        True, doc="Print the results of the benchmark function every time it is called"
    )

    clear_history: bool = param.Boolean(False, doc="Clear historical results")

    print_pandas: bool = param.Boolean(
        False, doc="Print a pandas summary of the results to the console."
    )

    print_xarray: bool = param.Boolean(
        False, doc="Print an xarray summary of the results to the console"
    )

    serve_pandas: bool = param.Boolean(
        False,
        doc="Serve a pandas summary on the results webpage.  If you have a large dataset consider setting this to false if the page loading is slow",
    )

    serve_pandas_flat: bool = param.Boolean(
        True,
        doc="Serve a flattened pandas summary on the results webpage.  If you have a large dataset consider setting this to false if the page loading is slow",
    )

    serve_xarray: bool = param.Boolean(
        False,
        doc="Serve an xarray summary on the results webpage. If you have a large dataset consider setting this to false if the page loading is slow",
    )

    auto_plot: bool = param.Boolean(
        True, doc=" Automatically dedeuce the best type of plot for the results."
    )

    raise_duplicate_exception: bool = param.Boolean(False, doc=" Used to debug unique plot names.")

    use_cache: bool = param.Boolean(
        False,
        doc="This is a benchmark level cache that stores the results of a fully completed benchmark. At the end of a benchmark the values are added to the cache but are not if the benchmark does not complete.  If you want to cache values during the benchmark you need to use the use_sample_cache option. Beware that depending on how you change code in the objective function, the cache could provide values that are not correct.",
    )

    clear_cache: bool = param.Boolean(
        False, doc=" Clear the cache of saved input->output mappings."
    )

    use_sample_cache: bool = param.Boolean(
        False,
        doc="If true, every time the benchmark function is called, bencher will check if that value has been calculated before and if so load the from the cache.  Note that the sample level cache is different from the benchmark level cache which only caches the aggregate of all the results at the end of the benchmark. This cache lets you stop a benchmark halfway through and continue. However, beware that depending on how you change code in the objective function, the cache could provide values that are not correct.",
    )

    only_hash_tag: bool = param.Boolean(
        False,
        doc="By default when checking if a sample has been calculated before it includes the hash of the greater benchmarking context.  This is safer because it means that data generated from one benchmark will not affect data from another benchmark.  However, if you are careful it can be more flexible to ignore which benchmark generated the data and only use the tag hash to check if that data has been calculated before. ie, you can create two benchmarks that sample a subset of the problem during exploration and give them the same tag, and then afterwards create a larger benchmark that covers the cases you already explored.  If this value is true, the combined benchmark will use any data from other benchmarks with the same tag.",
    )

    clear_sample_cache: bool = param.Boolean(
        False,
        doc="Clears the per-sample cache.  Use this if you get unexpected behavior.  The per_sample cache is tagged by the specific benchmark it was sampled from. So clearing the cache of one benchmark will not clear the cache of other benchmarks.",
    )

    overwrite_sample_cache: bool = param.Boolean(
        False,
        doc="If True, recalculate the value and overwrite the value stored in the sample cache",
    )

    only_plot: bool = param.Boolean(
        False, doc="Do not attempt to calculate benchmarks if no results are found in the cache"
    )

    use_holoview: bool = param.Boolean(False, doc="Use holoview for plotting")

    nightly: bool = param.Boolean(
        False, doc="Run a more extensive set of tests for a nightly benchmark"
    )

    time_event: str = param.String(
        None,
        doc="A string representation of a sequence over time, i.e. datetime, pull request number, or run number",
    )

    headless: bool = param.Boolean(False, doc="Run the benchmarks headlessly")

    render_plotly = param.Boolean(
        True,
        doc="Plotly and Bokeh don't play nicely together, so by default pre-render plotly figures to a non dynamic version so that bokeh plots correctly.  If you want interactive 3D graphs, set this to true but be aware that your 2D interactive graphs will probably stop working.",
    )

    level = param.Integer(
        default=0,
        bounds=[0, 12],
        doc="The level parameter is a method of defining the number samples to sweep over in a variable agnostic way, i.e you don't need to specify the number of samples for each variable as they are calculated dynamically from the sampling level.  See example_level.py for more information.",
    )

    run_tag = param.String(
        default="",
        doc="Define a tag for a run to isolate the results stored in the cache from other runs",
    )

    run_date = param.Date(
        default=datetime.now(),
        doc="The date the bench run was performed",
    )

    # parallel = param.Boolean(
    #     default=False,
    #     doc="Run the sweep in parallel.  Warning! You need to make sure your code is threadsafe before using this option",
    # )

    executor = param.Selector(
        objects=list(Executors),
        doc="The function can be run serially or in parallel with different futures executors",
    )

    plot_size = param.Integer(default=None, doc="Sets the width and height of the plot")
    plot_width = param.Integer(
        default=None,
        doc="Sets with width of the plots, this will override the plot_size parameter",
    )
    plot_height = param.Integer(
        default=None, doc="Sets the height of the plot, this will override the plot_size parameter"
    )

    @staticmethod
    def from_cmd_line() -> BenchRunCfg:  # pragma: no cover
        """create a BenchRunCfg by parsing command line arguments

        Returns:
            parsed args: parsed args
        """

        parser = argparse.ArgumentParser(description="benchmark")

        parser.add_argument(
            "--use-cache",
            action="store_true",
            help=BenchRunCfg.param.use_cache.doc,
        )

        parser.add_argument(
            "--only-plot",
            action="store_true",
            help=BenchRunCfg.param.only_plot.doc,
        )

        parser.add_argument(
            "--port",
            type=int,
            help=BenchRunCfg.param.port.doc,
        )

        parser.add_argument(
            "--nightly",
            type=lambda b: bool(str2bool(b)),
            nargs="?",
            const=False,
            default=False,
            help="turn on nightly benchmarking",
        )

        parser.add_argument(
            "--time_event",
            type=str,
            default=BenchRunCfg.param.time_event.default,
            help=BenchRunCfg.param.time_event.doc,
        )

        parser.add_argument(
            "--repeats",
            type=int,
            default=BenchRunCfg.param.repeats.default,
            help=BenchRunCfg.param.repeats.doc,
        )

        return BenchRunCfg(**vars(parser.parse_args()))


class BenchCfg(BenchRunCfg):
    """A class for storing the arguments to configure a benchmark protocol  If the inputs variables are the same the class should return the same hash and same filename.  This is so that historical data can be referenced and ensures that the generated plots are unique per benchmark"""

    input_vars = param.List(
        default=None,
        doc="A list of ParameterizedSweep variables to perform a parameter sweep over",
    )
    result_vars = param.List(
        default=None,
        doc="A list of ParameterizedSweep results collect and plot.",
    )

    const_vars = param.List(
        default=None,
        doc="Variables to keep constant but are different from the default value",
    )

    result_hmaps = param.List(default=None, doc="a list of holomap results")

    meta_vars = param.List(
        default=None,
        doc="Meta variables such as recording time and repeat id",
    )
    all_vars = param.List(
        default=None,
        doc="Stores a list of both the input_vars and meta_vars that are used to define a unique hash for the input",
    )
    iv_time = param.List(
        default=[],
        item_type=TimeSnapshot | TimeEvent,
        doc="A parameter to represent the sampling the same inputs over time as a scalar type",
    )

    iv_time_event = param.List(
        default=[],
        item_type=TimeEvent,
        doc="A parameter to represent the sampling the same inputs over time as a discrete type",
    )

    over_time: param.Boolean(
        False, doc="A parameter to control whether the function is sampled over time"
    )
    name: str = param.String(None, doc="The name of the benchmarkCfg")
    title: str = param.String(None, doc="The title of the benchmark")
    raise_duplicate_exception: str = param.Boolean(
        False, doc="Use this while debugging if filename generation is unique"
    )
    bench_name: str = param.String(
        None, doc="The name of the benchmark and the name of the save folder"
    )
    description: str = param.String(
        None,
        doc="A place to store a longer description of the function of the benchmark",
    )
    post_description: str = param.String(None, doc="A place to comment on the output of the graphs")

    has_results: bool = param.Boolean(
        False,
        doc="If this config has results, true, otherwise used to store titles and other bench metadata",
    )

    pass_repeat: bool = param.Boolean(
        False,
        doc="By default do not pass the kwarg 'repeat' to the benchmark function.  Set to true if you want the benchmark function to be passed the repeat number",
    )

    tag: str = param.String(
        "",
        doc="Use tags to group different benchmarks together. By default benchmarks are considered distinct from each other and are identified by the hash of their name and inputs, constants and results and tag, but you can optionally change the hash value to only depend on the tag.  This way you can have multiple unrelated benchmarks share values with each other based only on the tag value.",
    )

    hash_value: str = param.String(
        "",
        doc="store the hash value of the config to avoid having to hash multiple times",
    )

    plot_callbacks = param.List(
        None,
        doc="A callable that takes a BenchResult and returns panel representation of the results",
    )

    def __init__(self, **params):
        super().__init__(**params)
        self.plot_lib = None
        self.hmap_kdims = None
        self.iv_repeat = None

    def hash_persistent(self, include_repeats) -> str:
        """override the default hash function because the default hash function does not return the same value for the same inputs.  It references internal variables that are unique per instance of BenchCfg

        Args:
            include_repeats (bool) : by default include repeats as part of the hash except with using the sample cache
        """

        if include_repeats:
            # needed so that the historical xarray arrays are the same size
            repeats_hash = hash_sha1(self.repeats)
        else:
            repeats_hash = 0

        hash_val = hash_sha1(
            (
                hash_sha1(str(self.bench_name)),
                hash_sha1(str(self.title)),
                hash_sha1(self.over_time),
                repeats_hash,
                hash_sha1(self.tag),
            )
        )
        all_vars = self.input_vars + self.result_vars
        for v in all_vars:
            hash_val = hash_sha1((hash_val, v.hash_persistent()))

        for v in self.const_vars:
            hash_val = hash_sha1((v[0].hash_persistent(), hash_sha1(v[1])))

        return hash_val

    def inputs_as_str(self) -> List[str]:
        return [i.name for i in self.input_vars]

    def describe_sweep(self, width: int = 800, accordion=True) -> pn.pane.Markdown:
        """Produce a markdown summary of the sweep settings"""

        desc = pn.pane.Markdown(self.describe_benchmark(), width=width)
        if accordion:
            return pn.Accordion(("Data Collection Parameters", desc))
        return desc

    def describe_benchmark(self) -> str:
        """Generate a string summary of the inputs and results from a BenchCfg

        Returns:
            str: summary of BenchCfg
        """
        benchmark_sampling_str = ["```text"]
        benchmark_sampling_str.append("")

        benchmark_sampling_str.append("Input Variables:")
        for iv in self.input_vars:
            benchmark_sampling_str.extend(describe_variable(iv, True))

        if self.const_vars and (self.summarise_constant_inputs):
            benchmark_sampling_str.append("\nConstants:")
            for cv in self.const_vars:
                benchmark_sampling_str.extend(describe_variable(cv[0], False, cv[1]))

        benchmark_sampling_str.append("\nResult Variables:")
        for rv in self.result_vars:
            benchmark_sampling_str.extend(describe_variable(rv, False))

        benchmark_sampling_str.append("\nMeta Variables:")
        benchmark_sampling_str.append(f"    run date: {self.run_date}")
        if self.run_tag:
            benchmark_sampling_str.append(f"    run tag: {self.run_tag}")
        if self.level is not None:
            benchmark_sampling_str.append(f"    bench level: {self.level}")
        benchmark_sampling_str.append(f"    use_cache: {self.use_cache}")
        benchmark_sampling_str.append(f"    use_sample_cache: {self.use_sample_cache}")
        benchmark_sampling_str.append(f"    only_hash_tag: {self.only_hash_tag}")
        benchmark_sampling_str.append(f"    executor: {self.executor}")

        for mv in self.meta_vars:
            benchmark_sampling_str.extend(describe_variable(mv, True))

        benchmark_sampling_str.append("```")

        benchmark_sampling_str = "\n".join(benchmark_sampling_str)
        return benchmark_sampling_str

    def to_title(self, panel_name: str = None) -> pn.pane.Markdown:
        if panel_name is None:
            panel_name = self.title
        return pn.pane.Markdown(f"# {self.title}", name=panel_name)

    def to_description(self, width: int = 800) -> pn.pane.Markdown:
        return pn.pane.Markdown(f"{self.description}", width=width)

    def to_post_description(self, width: int = 800) -> pn.pane.Markdown:
        return pn.pane.Markdown(f"{self.post_description}", width=width)

    def to_sweep_summary(
        self,
        name=None,
        description=True,
        describe_sweep=True,
        results_suffix=True,
        title: bool = True,
    ) -> pn.pane.Markdown:
        """Produce panel output summarising the title, description and sweep setting"""
        if name is None:
            if title:
                name = self.title
            else:
                name = "Data Collection Parameters"
        col = pn.Column(name=name)
        if title:
            col.append(self.to_title())
        if self.description is not None and description:
            col.append(self.to_description())
        if describe_sweep:
            col.append(self.describe_sweep())
        if results_suffix:
            col.append(pn.pane.Markdown("## Results:"))
        return col

    def optuna_targets(self, as_var=False) -> List[str]:
        targets = []
        for rv in self.result_vars:
            if hasattr(rv, "direction") and rv.direction != OptDir.none:
                if as_var:
                    targets.append(rv)
                else:
                    targets.append(rv.name)
        return targets


class DimsCfg:
    """A class to store data about the sampling and result dimensions"""

    def __init__(self, bench_cfg: BenchCfg) -> None:
        self.dims_name = [i.name for i in bench_cfg.all_vars]

        self.dim_ranges = []
        self.dim_ranges = [i.values() for i in bench_cfg.all_vars]
        self.dims_size = [len(p) for p in self.dim_ranges]
        self.dim_ranges_index = [list(range(i)) for i in self.dims_size]
        self.dim_ranges_str = [f"{s}\n" for s in self.dim_ranges]
        self.coords = dict(zip(self.dims_name, self.dim_ranges))

        logging.debug(f"dims_name: {self.dims_name}")
        logging.debug(f"dim_ranges {self.dim_ranges_str}")
        logging.debug(f"dim_ranges_index {self.dim_ranges_index}")
        logging.debug(f"coords: {self.coords}")



import holoviews as hv
import numpy as np
import panel as pn

# hv.extension("bokeh", "plotly")
backend1 = "plotly"
# backend1 = "matplotlib"
backend2 = "bokeh"

# matplotlib.use("agg")

# hv.extension(backend1, backend2)
hv.extension(backend2, backend1)

# hv.extension(backend1)

# opts.defaults(opts.Surface(backend=backend1), opts.HeatMap(backend=backend2))

# hv.output(opts.Surface(backend=backend1), opts.HeatMap(backend=backend2))
# hv.output(surface.opts(fig_size=200, backend='matplotlib'), backend='matplotlib')

main = pn.Row()

# hv.output(backend="plotly")

dat = np.ones([10, 10])
# main.append(hv.Surface(dat))

surf = hv.Surface(dat)
main.append(hv.render(surf, backend=backend1))


# hv.output(backend="bokeh")
heat = hv.HeatMap(dat)
heat *= hv.Text(0, 0, "I only work with bokeh and matplotlib")
main.append(heat)

main.show()



from __future__ import annotations
from strenum import StrEnum
from typing import Any
import importlib
from abc import abstractmethod
from dataclasses import dataclass
from enum import auto


class ClassEnum(StrEnum):
    """A ClassEnum is a pattern to make it easier to create factory a factory method that converts from an enum to a corresponding class.  Subclasses should implement to_class(enum_instance:EnumType) which takes an enum returns the corresponding instance of that class."""

    @classmethod
    def to_class_generic(cls, module_import: str, class_name: str) -> Any:
        """Create an instance of the class referred to by this enum

        Returns:
            Any: instance of the class
        """

        class_def = getattr(importlib.import_module(module_import), class_name)
        return class_def()

    @classmethod
    @abstractmethod
    def to_class(cls, enum_val: ClassEnum) -> Any:
        """Subclasses should overrides this method to take  an enum returns the corresponding instance of that class."""
        raise NotImplementedError()


@dataclass
class BaseClass:
    baseclassname: str = "class0"


@dataclass
class Class1(BaseClass):
    classname: str = "class1"


@dataclass
class Class2(BaseClass):
    classname: str = "class2"


class ExampleEnum(ClassEnum):
    Class1 = auto()
    Class2 = auto()

    @classmethod
    def to_class(cls, enum_val: ExampleEnum) -> BaseClass:
        return cls.to_class_generic("bencher.class_enum", enum_val)



from typing import Optional
import panel as pn
from param import Parameter
from bencher.results.bench_result_base import BenchResultBase, ReduceType


from functools import partial
import holoviews as hv
from bencher.variables.results import (
    PANEL_TYPES,
)


class DataSetResult(BenchResultBase):
    def to_dataset1(
        self,
        result_var: Parameter = None,
        hv_dataset=None,
        target_dimension: int = 0,
        container=None,
        level: int = None,
        **kwargs,
    ) -> Optional[pn.pane.panel]:
        if hv_dataset is None:
            hv_dataset = self.to_hv_dataset(ReduceType.SQUEEZE, level=level)
        elif not isinstance(hv_dataset, hv.Dataset):
            hv_dataset = hv.Dataset(hv_dataset)
        return self.map_plot_panes(
            partial(self.ds_to_container, container=container),
            hv_dataset=hv_dataset,
            target_dimension=target_dimension,
            result_var=result_var,
            result_types=PANEL_TYPES,
            **kwargs,
        )


# class DataSetResult(BenchResultBase):

#     def to_datatset(
#         self,
#         result_var: Parameter = None,
#         result_types=(ResultDataSet,),
#         pane_collection: pn.pane = None,
#         time_sequence_dimension=0,
#         target_duration: float = None,
#         **kwargs,
#     ) -> Optional[pn.panel]:
#         """Returns the results compiled into a video

#         Args:
#             result_var (Parameter, optional): The result var to plot. Defaults to None.
#             result_types (tuple, optional): The types of result var to convert to video. Defaults to (ResultDataSet,).
#             collection (pn.pane, optional): If there are multiple results, use this collection to stack them. Defaults to pn.Row().

#         Returns:
#             Optional[pn.panel]: a panel pane with a video of all results concatenated together
#         """
#         plot_filter = PlotFilter(
#             float_range=VarRange(0, None),
#             cat_range=VarRange(0, None),
#             panel_range=VarRange(1, None),
#             input_range=VarRange(0, None),
#         )
#         matches_res = plot_filter.matches_result(
#             self.plt_cnt_cfg, callable_name(self.to_video_grid_ds)
#         )

#         if pane_collection is None:
#             pane_collection = pn.Row()

#         if matches_res.overall:
#             ds = self.to_dataset(ReduceType.SQUEEZE)
#             for rv in self.get_results_var_list(result_var):
#                 if isinstance(rv, result_types):
#                     pane_collection.append(
#                         self.to_video_grid_ds(
#                             ds,
#                             rv,
#                             time_sequence_dimension=time_sequence_dimension,
#                             target_duration=target_duration,
#                             **kwargs,
#                         )
#                     )
#             return pane_collection
#         return matches_res.to_panel()

#     def to_video_grid_ds(
#         self,
#         dataset: xr.Dataset,
#         result_var: Parameter,
#         reverse=True,
#         time_sequence_dimension=0,
#         video_controls: VideoControls = None,
#         target_duration: float = None,
#         **kwargs,
#     ):
#         cvc = self._to_video_panes_ds(
#             dataset,
#             self.plot_cb,
#             target_dimension=0,
#             horizontal=True,
#             compose_method=ComposeType.right,
#             time_sequence_dimension=time_sequence_dimension,
#             result_var=result_var,
#             final=True,
#             reverse=reverse,
#             target_duration=target_duration,
#             **kwargs,
#         )

#         filename = VideoWriter().write_video_raw(cvc)

#         if filename is not None:
#             if video_controls is None:
#                 video_controls = VideoControls()
#             return video_controls.video_container(
#                 filename, width=kwargs.get("width", None), height=kwargs.get("height", None)
#             )
#         return None

#     def plot_cb(self, dataset, result_var, **kwargs):
#         val = self.ds_to_container(dataset, result_var, container=None, **kwargs)
#         return val

#     def dataset_to_compose_list(
#         self,
#         dataset: xr.Dataset,
#         first_compose_method: ComposeType = ComposeType.down,
#         time_sequence_dimension: int = 0,
#     ) -> List[ComposeType]:
#         """ "Given a dataset, chose an order for composing the results.  By default will flip between right and down and the last dimension will be a time sequence.

#         Args:
#             dataset (xr.Dataset): the dataset to render
#             first_compose_method (ComposeType, optional): the direction of the first composition method. Defaults to ComposeType.right.
#             time_sequence_dimension (int, optional): The dimension to start time sequencing instead of composing in space. Defaults to 0.

#         Returns:
#             List[ComposeType]: A list of composition methods for composing the dataset result
#         """

#         num_dims = len(dataset.sizes)
#         if time_sequence_dimension == -1:  # use time sequence for everything
#             compose_method_list = [ComposeType.sequence] * (num_dims + 1)
#         else:
#             compose_method_list = [first_compose_method]
#             compose_method_list.extend(
#                 ComposeType.flip(compose_method_list[-1]) for _ in range(num_dims - 1)
#             )
#             compose_method_list.append(ComposeType.sequence)

#             for i in range(min(len(compose_method_list), time_sequence_dimension + 1)):
#                 compose_method_list[i] = ComposeType.sequence

#         return compose_method_list

#     def _to_video_panes_ds(
#         self,
#         dataset: xr.Dataset,
#         plot_callback: callable = None,
#         target_dimension=0,
#         compose_method=ComposeType.right,
#         compose_method_list=None,
#         result_var=None,
#         time_sequence_dimension=0,
#         root_dimensions=None,
#         reverse=False,
#         target_duration: float = None,
#         **kwargs,
#     ) -> pn.panel:
#         num_dims = len(dataset.sizes)
#         dims = list(d for d in dataset.sizes)
#         if reverse:
#             dims = list(reversed(dims))

#         if root_dimensions is None:
#             root_dimensions = num_dims

#         if compose_method_list is None:
#             compose_method_list = self.dataset_to_compose_list(
#                 dataset, compose_method, time_sequence_dimension=time_sequence_dimension
#             )

#             # print(compose_method_list)

#         compose_method_list_pop = deepcopy(compose_method_list)
#         if len(compose_method_list_pop) > 1:
#             compose_method = compose_method_list_pop.pop()

#         if num_dims > (target_dimension) and num_dims != 0:
#             selected_dim = dims[-1]
#             outer_container = ComposableContainerVideo()
#             for i in range(dataset.sizes[selected_dim]):
#                 sliced = dataset.isel({selected_dim: i})
#                 label_val = sliced.coords[selected_dim].values.item()
#                 inner_container = ComposableContainerVideo()

#                 panes = self._to_video_panes_ds(
#                     sliced,
#                     plot_callback=plot_callback,
#                     target_dimension=target_dimension,
#                     compose_method_list=compose_method_list_pop,
#                     result_var=result_var,
#                     root_dimensions=root_dimensions,
#                     time_sequence_dimension=time_sequence_dimension,
#                 )
#                 inner_container.append(panes)

#                 rendered = inner_container.render(
#                     RenderCfg(
#                         var_name=selected_dim,
#                         var_value=label_val,
#                         compose_method=compose_method,
#                         duration=target_duration,
#                     )
#                 )
#                 outer_container.append(rendered)
#             return outer_container.render(
#                 RenderCfg(
#                     compose_method=compose_method,
#                     duration=target_duration,
#                     background_col=color_tuple_to_255(int_to_col(num_dims - 2, 0.05, 1.0)),
#                     # background_col= (255,0,0),
#                 )
#             )
#         return plot_callback(dataset=dataset, result_var=result_var, **kwargs)



from __future__ import annotations
import numpy as np
from copy import deepcopy
from pathlib import Path
from dataclasses import dataclass
from moviepy.editor import (
    ImageClip,
    CompositeVideoClip,
    clips_array,
    concatenate_videoclips,
    VideoClip,
    VideoFileClip,
)
from moviepy.video.fx.margin import margin

from bencher.results.composable_container.composable_container_base import (
    ComposableContainerBase,
    ComposeType,
)
from bencher.video_writer import VideoWriter


@dataclass()
class RenderCfg:
    compose_method: ComposeType = ComposeType.sequence
    var_name: str = None
    var_value: str = None
    background_col: tuple[int, int, int] = (255, 255, 255)
    duration: float = 10.0
    duration_target: bool = True
    min_frame_duration: float = 1.0 / 30
    max_frame_duration: float = 2.0
    margin: int = 0


@dataclass
class ComposableContainerVideo(ComposableContainerBase):
    def append(self, obj: VideoClip | ImageClip | str | np.ndarray) -> None:
        """Appends an image or video to the container

        Args:
            obj (VideoClip | ImageClip | str | np.ndarray): Any representation of an image or video

        Raises:
            RuntimeWarning: if file format is not recognised
        """

        # print(f"append obj: {type(obj)}, {obj}")
        if obj is not None:
            if isinstance(obj, VideoClip):
                self.container.append(obj)
            elif isinstance(obj, ComposableContainerVideo):
                self.container.append(obj.render())
            elif isinstance(obj, np.ndarray):
                self.container.append(ImageClip(obj))
            else:
                path = Path(obj)
                extension = str.lower(path.suffix)
                if extension in [".jpg", ".jepg", ".png"]:
                    self.container.append(ImageClip(obj))
                elif extension in [".mpeg", ".mpg", ".mp4", ".webm"]:
                    # print(obj)
                    self.container.append(VideoFileClip(obj))
                else:
                    raise RuntimeWarning(f"unsupported filetype {extension}")
        else:
            raise RuntimeWarning("No data passed to ComposableContainerVideo.append()")

    def calculate_duration(self, frames, render_cfg: RenderCfg):
        if render_cfg.duration_target:
            # calculate duration based on fps constraints
            duration = 10.0 if render_cfg.duration is None else render_cfg.duration
            frame_duration = duration / frames
            if render_cfg.min_frame_duration is not None:
                frame_duration = max(frame_duration, render_cfg.min_frame_duration)
            if render_cfg.max_frame_duration is not None:
                frame_duration = min(frame_duration, render_cfg.max_frame_duration)
            duration = frame_duration * frames
        else:
            duration = render_cfg.duration
            frame_duration = duration / float(frames)

        print("max_frame_duration", render_cfg.max_frame_duration)
        print("DURATION", duration)

        return duration, frame_duration

    def render(self, render_cfg: RenderCfg = None, **kwargs) -> CompositeVideoClip:
        """Composes the images/videos into a single image/video based on the type of compose method

        Args:
            compose_method (ComposeType, optional): optionally override the default compose type. Defaults to None.

        Returns:
            CompositeVideoClip: A composite video clip containing the images/videos added via append()
        """
        if render_cfg is None:
            render_cfg = RenderCfg(**kwargs)

        print("rc", render_cfg)
        _, frame_duration = self.calculate_duration(float(len(self.container)), render_cfg)
        out = None
        print(f"using compose type{render_cfg.compose_method}")
        max_duration = 0.0

        for i in range(len(self.container)):
            if self.container[i].duration is None:
                self.container[i].duration = frame_duration
            max_duration = max(max_duration, self.container[i].duration)
        match render_cfg.compose_method:
            case ComposeType.right | ComposeType.down:
                for i in range(len(self.container)):
                    self.container[i] = self.extend_clip(self.container[i], max_duration)
                    self.container[i] = margin(
                        self.container[i], top=render_cfg.margin, color=render_cfg.background_col
                    )

                if render_cfg.compose_method == ComposeType.right:
                    clips = [self.container]
                else:
                    clips = [[c] for c in self.container]
                out = clips_array(clips, bg_color=render_cfg.background_col)
                if out.duration is None:
                    out.duration = max_duration
            case ComposeType.sequence:
                out = concatenate_videoclips(
                    self.container, bg_color=render_cfg.background_col, method="compose"
                )
            # case ComposeType.overlay:
            #     for i in range(len(self.container)):
            #         self.container[i].alpha = 1./len(self.container)
            #     out = CompositeVideoClip(self.container, bg_color=render_args.background_col)
            #     out.duration = fps
            case _:
                raise RuntimeError("This compose type is not supported")

        label = self.label_formatter(render_cfg.var_name, render_cfg.var_value)
        if label is not None:
            # print("adding label")
            label = ImageClip(
                np.array(VideoWriter.create_label(label, color=render_cfg.background_col))
            )
            label.duration = out.duration
            label_compose = ComposeType.down
            if render_cfg.compose_method == ComposeType.down:
                label_compose = ComposeType.right
            con2 = ComposableContainerVideo()
            con2.append(label)
            con2.append(out)
            return con2.render(
                RenderCfg(
                    background_col=render_cfg.background_col,
                    compose_method=label_compose,
                    duration=out.duration,
                    duration_target=False,  # want exact duration
                )
            )
        return out

    def to_video(
        self,
        render_args: RenderCfg = None,
    ) -> str:
        """Returns the composite video clip as a webm file path

        Returns:
            str: webm filepath
        """
        return VideoWriter().write_video_raw(self.render(render_args))

    def deep(self):
        return deepcopy(self)

    def extend_clip(self, clip: VideoClip, desired_duration: float):
        if clip.duration < desired_duration:
            return concatenate_videoclips(
                [
                    clip,
                    ImageClip(
                        clip.get_frame(clip.duration), duration=desired_duration - clip.duration
                    ),
                ]
            )
        return clip



# pylint: skip-file  #this is experimental still


from bokeh.models import HoverTool
import pandas as pd
from hvplot import pandas  # noqa
import holoviews as hv
import panel as pn

datadict = dict(
    x=[1, 5],
    y=[4, 10],
    img=[
        "https://raw.githubusercontent.com/holoviz/panel/master/doc/_static/logo_horizontal.png",
        "https://raw.githubusercontent.com/holoviz/panel/master/doc/_static/logo_stacked.png",
    ],
)


hover = HoverTool(
    tooltips="""
    <div>
        <div>
            <img src="@img" width=100 style="float: left; margin: 0px 15px 15px 0px; border="2"></img>
        <div>
            <span style="font-size: 15px;">Location</span>
            <span style="font-size: 10px; color: #696;">(@x, @y)</span>
        </div>
    </div>

"""
)

df_test = pd.DataFrame.from_dict(datadict)


hvds = hv.Dataset(df_test)

pt = hv.Scatter(hvds, kdims=["x"], vdims=["y"], hover_cols=["img"], tools=[hover])

pn.Row(pt).show()

# pt =df_test.hvplot.scatter(x="x", y="y", hover_cols=["img"], tools=[hover])

# pt.show()



import logging
from typing import Callable
import os
from pathlib import Path
import tempfile
from threading import Thread
from dataclasses import dataclass

import panel as pn
from bencher.results.bench_result import BenchResult
from bencher.bench_plot_server import BenchPlotServer
from bencher.bench_cfg import BenchRunCfg


@dataclass
class GithubPagesCfg:
    github_user: str
    repo_name: str
    folder_name: str = "report"
    branch_name: str = "gh-pages"


class BenchReport(BenchPlotServer):
    def __init__(
        self,
        bench_name: str = None,
    ) -> None:
        self.bench_name = bench_name
        self.pane = pn.Tabs(tabs_location="left", name=self.bench_name)

    def append_title(self, title: str, new_tab: bool = True):
        if new_tab:
            return self.append_tab(pn.pane.Markdown(f"# {title}", name=title), title)
        return self.append_markdown(f"# {title}", title)

    def append_markdown(self, markdown: str, name=None, width=800, **kwargs) -> pn.pane.Markdown:
        if name is None:
            name = markdown
        md = pn.pane.Markdown(markdown, name=name, width=width, **kwargs)
        self.append(md, name)
        return md

    def append(self, pane: pn.panel, name: str = None) -> None:
        if len(self.pane) == 0:
            if name is None:
                name = pane.name
            self.append_tab(pane, name)
        else:
            self.pane[-1].append(pane)

    def append_col(self, pane: pn.panel, name: str = None) -> None:
        if name is not None:
            col = pn.Column(pane, name=name)
        else:
            col = pn.Column(pane, name=pane.name)
        self.pane.append(col)

    def append_result(self, bench_res: BenchResult) -> None:
        self.append_tab(bench_res.plot(), bench_res.bench_cfg.title)

    def append_tab(self, pane: pn.panel, name: str = None) -> None:
        if pane is not None:
            if name is None:
                name = pane.name
            self.pane.append(pn.Column(pane, name=name))

    def save_index(self, directory="", filename="index.html") -> Path:
        """Saves the result to index.html in the root folder so that it can be displayed by github pages.

        Returns:
            Path: save path
        """
        return self.save(directory, filename, False)

    def save(
        self,
        directory: str | Path = "cachedir",
        filename: str = None,
        in_html_folder: bool = True,
        **kwargs,
    ) -> Path:
        """Save the result to a html file.  Note that dynamic content will not work.  by passing save(__file__) the html output will be saved in the same folder as the source code in a html subfolder.

        Args:
            directory (str | Path, optional): base folder to save to. Defaults to "cachedir" which should be ignored by git.
            filename (str, optional): The name of the html file. Defaults to the name of the benchmark
            in_html_folder (bool, optional): Put the saved files in a html subfolder to help keep the results separate from source code. Defaults to True.

        Returns:
            Path: the save path
        """

        if filename is None:
            filename = f"{self.bench_name}.html"

        base_path = Path(directory)

        if in_html_folder:
            base_path /= "html"

        logging.info(f"creating dir {base_path.absolute()}")
        os.makedirs(base_path.absolute(), exist_ok=True)

        base_path = base_path / filename

        logging.info(f"saving html output to: {base_path.absolute()}")

        self.pane.save(filename=base_path, progress=True, embed=True, **kwargs)
        return base_path

    def show(self, run_cfg: BenchRunCfg = None) -> Thread:  # pragma: no cover
        """Launches a webserver with plots of the benchmark results, blocking

        Args:
            run_cfg (BenchRunCfg, optional): Options for the webserve such as the port. Defaults to None.

        """
        if run_cfg is None:
            run_cfg = BenchRunCfg()

        return BenchPlotServer().plot_server(self.bench_name, run_cfg, self.pane)

    def publish_gh_pages(
        self,
        github_user: str,
        repo_name: str,
        folder_name: str = "report",
        branch_name: str = "gh-pages",
    ) -> str:  # pragma: no cover
        remote = f"https://github.com/{github_user}/{repo_name}.git"
        publish_url = f"https://{github_user}.github.io/{repo_name}/{folder_name}"

        with tempfile.TemporaryDirectory() as td:
            directory = td
            report_path = self.save(
                directory + f"/{folder_name}/",
                filename="index.html",
                in_html_folder=False,
            )
            logging.info(f"created report at: {report_path.absolute()}")

            cd_dir = f"cd {directory} &&"
            # TODO DON'T OVERWRITE EVERYTHING
            os.system(f"{cd_dir} git init")
            os.system(f"{cd_dir} git checkout -b {branch_name}")
            os.system(f"{cd_dir} git add {folder_name}/index.html")
            os.system(f'{cd_dir} git commit -m "publish {branch_name}"')
            os.system(f"{cd_dir} git remote add origin {remote}")
            os.system(f"{cd_dir} git push --set-upstream origin {branch_name} -f")

        logging.info("Published report @")
        logging.info(publish_url)

        return publish_url

    def publish(
        self, remote_callback: Callable, branch_name: str = None, debug: bool = False
    ) -> str:  # pragma: no cover
        """Publish the results as an html file by committing it to the bench_results branch in the current repo. If you have set up your repo with github pages or equivalent then the html file will be served as a viewable webpage.  This is an example of a callable to publish on github pages:

        .. code-block:: python

            def publish_args(branch_name) -> Tuple[str, str]:
                return (
                    "https://github.com/dyson-ai/bencher.git",
                    f"https://github.com/dyson-ai/bencher/blob/{branch_name}")


        Args:
            remote (Callable): A function the returns a tuple of the publishing urls. It must follow the signature def publish_args(branch_name) -> Tuple[str, str].  The first url is the git repo name, the second url needs to match the format for viewable html pages on your git provider.  The second url can use the argument branch_name to point to the report on a specified branch.

        Returns:
            str: the url of the published report
        """

        if branch_name is None:
            branch_name = self.bench_name
        branch_name += "_debug" if debug else ""

        remote, publish_url = remote_callback(branch_name)

        with tempfile.TemporaryDirectory() as td:
            directory = td
            report_path = self.save(directory, filename="index.html", in_html_folder=False)
            logging.info(f"created report at: {report_path.absolute()}")

            cd_dir = f"cd {directory} &&"

            os.system(f"{cd_dir} git init")
            os.system(f"{cd_dir} git checkout -b {branch_name}")
            os.system(f"{cd_dir} git add index.html")
            os.system(f'{cd_dir} git commit -m "publish {branch_name}"')
            os.system(f"{cd_dir} git remote add origin {remote}")
            os.system(f"{cd_dir} git push --set-upstream origin {branch_name} -f")

        logging.info("Published report @")
        logging.info(publish_url)

        return publish_url
